!=======================================================================
!
! SUBROUTINE: bsstep, mmid, pzextr
!
! PURPOSE   : Bulirsch-Stoer integrator with adaptive stepsize,
!             using modified midpoint method (subroutine mmid)
!             and polynomial extrapolation (subroutine pzextr).
!
! SOURCE    : Numerical Recipes, Press et al. 2nd ed. 1992
!
!=======================================================================


!=======================================================================
module my_type
integer, parameter :: i4b = selected_int_kind(9)      !integer*4
integer, parameter :: dp = selected_real_kind(15,307) !real*8
end module my_type
!=======================================================================


!=======================================================================
subroutine bsstep(nvar   ,                  &
                  x      , htry   , eps   , &
                  y      , dydx   , yscal , &
                  narg   , argarr ,         &
                  hdid   , hnext  , &
                  logunit)

  use my_type
  implicit none

  integer(i4b) :: nvar, narg, logunit
  real(dp)     :: x, htry, eps, hdid, hnext
  real(dp)     :: y(nvar), dydx(nvar), yscal(nvar)
  real(dp)     :: argarr(narg)

  integer(i4b), parameter :: kmaxx = 8, imax = kmaxx + 1
  real(dp),     parameter :: safe1 = .25d0,  safe2 = .7d0, &
                             redmax = 1.d-5, redmin = .7d0, &
                             scalmx = .1d0,  tiny = 1.d-30

  integer(i4b) :: i, iq, k, kk, km
  integer(i4b), save :: kmax, kopt, nseq(imax)
  real(dp) :: eps1, errmax, fact, h, red, scale, &
              work, wrkmin, xest, err(kmaxx)
  real(dp), save :: a(imax), alf(kmaxx, kmaxx), epsold, xnew
  real(dp), allocatable, dimension(:) :: yerr, ysav, yseq

  logical       :: reduct
  logical, save :: first

  external derivs

  data first / .true. /, epsold / -1.d0 /
  data nseq / 2, 4, 6, 8, 10, 12, 14, 16, 18 /

  allocate(yerr(nvar),ysav(nvar),yseq(nvar))

  if (eps.ne.epsold) then
    hnext = -1.d29
    xnew =  -1.d29
    eps1 = safe1 * eps
    a(1) = nseq(1) + 1
    do k = 1, kmaxx
      a(k + 1) = a(k) + nseq(k + 1)
    end do

    do iq = 2, kmaxx
      do k = 1, iq - 1
        alf(k, iq) = eps1** ( (a(k + 1) - a(iq + 1) ) / &
                     ( (a(iq + 1) - a(1) + 1.d0) * (2 * k + 1) ) )
      end do
    end do

    epsold = eps

    do kopt = 2, kmaxx - 1
      if (a(kopt + 1) .gt. a(kopt) * alf(kopt - 1, kopt) ) goto 1
    end do
1   kmax = kopt
  endif

  h = htry
  ysav = y

  if (h.ne.hnext.or.x.ne.xnew) then
    first = .true.
    kopt = kmax
  endif

  reduct = .false.
2 do k = 1, kmax
    xnew = x + h
!   if (xnew.eq.x) pause 'step size underflow in bsstep'

    call mmid (ysav, dydx, nvar, x, h, nseq(k), yseq, derivs, narg, argarr)

    xest = (h / nseq(k) ) **2

    call pzextr (k, xest, yseq, y, yerr, nvar)

    if (k.ne.1) then
      errmax = tiny
      do i = 1, nvar
        errmax = max (errmax, abs (yerr(i) / yscal(i) ) )
      end do
      errmax = errmax / eps
      km = k - 1
      err(km) = (errmax / safe1) ** (1.d0 / (2 * km + 1) )
    endif

    if (k.ne.1.and. (k.ge.kopt - 1.or.first) ) then
      if (errmax.lt.1.) goto 4
      if (k.eq.kmax.or.k.eq.kopt + 1) then
        red = safe2 / err(km)
        goto 3
      elseif (k.eq.kopt) then
        if (alf(kopt - 1, kopt) .lt.err(km) ) then
          red = 1. / err(km)
          goto 3
        endif
      elseif (kopt.eq.kmax) then
        if (alf(km, kmax - 1) .lt.err(km) ) then
          red = alf(km, kmax - 1) * safe2 / err(km)
          goto 3
        endif
      elseif (alf(km, kopt) .lt.err(km) ) then
        red = alf(km, kopt - 1) / err(km)
        goto 3
      endif
    endif
  end do

3 red = min (red, redmin)
  red = max (red, redmax)
  h = h * red
  reduct = .true.
  goto 2

4 x = xnew
  hdid = h
  first = .false.
  wrkmin = 1.d35

  do kk = 1, km
    fact = max (err(kk), scalmx)
    work = fact * a(kk + 1)
    if (work.lt.wrkmin) then
      scale = fact
      wrkmin = work
      kopt = kk + 1
    endif
  end do

  hnext = h / scale
  if (kopt.ge.k.and.kopt.ne.kmax.and..not.reduct) then
    fact = max (scale / alf(kopt - 1, kopt), scalmx)
    if (a(kopt + 1) * fact.le.wrkmin) then
      hnext = h / fact
      kopt = kopt + 1
    endif
  endif

  deallocate(yerr,ysav,yseq)

  return
end subroutine bsstep
!=======================================================================


!=======================================================================
subroutine mmid (y, dydx, nvar, xs, htot, nstep, yout, derivs, narg, argarr)

  use my_type
  implicit none

  integer(i4b) :: nstep, nvar, narg
  real(dp)     :: htot, xs, dydx(nvar), y(nvar), yout(nvar), argarr(narg)
  external     :: derivs

  integer(i4b) :: i, n
  real(dp)     :: h, h2, swap, x

  real(dp), allocatable, dimension(:) ::  ym, yn

  allocate(ym(nvar), yn(nvar))

  h = htot / nstep

  do i = 1, nvar
    ym(i) = y(i)
    yn(i) = y(i) + h * dydx(i)
  enddo

  x = xs + h

  call derivs(nvar, x, yn, narg, argarr, yout)

  h2 = 2. * h

  do n = 2, nstep
    do i = 1, nvar
      swap = ym(i) + h2 * yout(i)
      ym(i) = yn(i)
      yn(i) = swap
    enddo
    x = x + h
    call derivs(nvar, x, yn, narg, argarr, yout)
  enddo

  do i = 1, nvar
    yout(i) = 0.5 * (ym(i) + yn(i) + h * yout(i))
  enddo

  deallocate(yn, ym)

  return
end subroutine mmid
!=======================================================================


!=======================================================================
subroutine pzextr (iest, xest, yest, yz, dy, nv)

  use my_type
  implicit none

  integer(i4b) :: iest, nv
  real(dp)     :: xest, yest(nv), yz(nv), dy(nv)

  integer(i4b), parameter :: imax = 13
  integer(i4b)   :: j, k1
  real(dp)       :: delta, f1, f2, q
  real(dp), save :: x(imax)

  real(dp), allocatable, dimension(:)         :: d
  real(dp), allocatable, dimension(:,:), save :: qcol

  allocate(d(nv), qcol(nv,imax))

  x(iest) = xest
  do j = 1, nv
    dy(j) = yest(j)
    yz(j) = yest(j)
  enddo

  if (iest.eq.1) then
    do j = 1, nv
      qcol(j, 1) = yest(j)
    enddo
  else
    do j = 1, nv
      d(j) = yest(j)
    end do
    do k1 = 1, iest - 1
      delta = 1. / (x(iest - k1) - xest)
      f1 = xest * delta
      f2 = x(iest - k1) * delta
      do j = 1, nv
        q = qcol(j, k1)
        qcol(j, k1) = dy(j)
        delta = d(j) - q
        dy(j) = f1 * delta
        d(j)  = f2 * delta
        yz(j) = yz(j) + dy(j)
      enddo
    enddo
    do j = 1, nv
      qcol(j, iest) = dy(j)
    end do
  endif

  deallocate(qcol,d)

  return
end subroutine pzextr
!=======================================================================

