!=======================================================================
!
! PROGRAM : frame_error.F
!
! PURPOSE : Frame for integrator of coupled ODEs using odeint,
!           to calculate global error of problem 1 as a function of h
!           for Euler and RK4
!
! SOURCE  :
!
! INPUT   : frame.inp  : integrator control and basic variables
!           problem.inp: problem specific parameters
!                        (constant factors for derivatives, etc.)
!
! OUTPUT  : xp     : intermediate values x
!           yp     : intermediate values y
!           ystart : end values
!
! EXTERNAL: subroutine derivs(nvar, x, y, narg, argarr, dydx)
!
! STRUCTURE:
!           frame [main program] calls
!             setproblem   [specifies problem number and the number of ODEs]
!             initproblem  [specifies - depending on problem number -
!                           constant factors and initial condition]
!             odeint       [integrator driver] calls
!               derivs     [derivatives - first call]
!               eulerstep || rk4 || rkqs || stiff [the actual integrators]
!
!             The integrators may call derivs or other helper routines.
!
! USAGE   : * Derivatives (right-hand side of ODEs) need to be supplied
!             by user in routine derivs (interface see above).
!           * Problem control (set of ODEs to be integrated):
!             Currently, there are three problems implemented:
!               iprob = 1: y' = ay
!               iprob = 2: stiff equation (1), NEEDS TO BE COMPLETED
!               iprob = 3: stiff equation (2)
!
!             For definition of a new problem, take next iprob and
!             change the following routines:
!               setproblem.F : set nvar and narg appropriate to ODEs
!               initproblem.F: set ystart and argarr to appropriate values.
!                              The 1st element of argarr needs to contain iprob,
!                              for use in derivs.F
!               derivs.F     : code the right-hand side of the ODEs (the derivatives)
!                              using the constant coefficients in argarr.
!
!=======================================================================

program frame_error

  implicit none
  integer, parameter :: i8 = selected_int_kind(18)

  integer(kind=i8) :: maxstep
  integer :: logunit, dmpunit, inpunit
  integer :: iprob, nvar, narg, iwhich, kmax, kcount, nok, nbad, done, i, it
  double precision :: x1, x2, eps, h1, hmin, dxsav, h, yexact, dxsav_orig, stepfac

  double precision, allocatable :: xp    (:)   ! storage for the independent variable (1:kmax)
  double precision, allocatable :: yp    (:,:) ! storage for the dependent variables (1:kmax,1:nvar)
  double precision, allocatable :: argarr(:)   ! argument array for derivs
  double precision, allocatable :: ystart(:)   ! initial conditions for dependent variables at x1

  character(6)  :: label ! label for dump file
  character(15) :: fnm   ! filename for dump file

!------------------------------------------------------------------

! define the file units
  dmpunit = 40
  inpunit =  1

! read parameters from file frame.inp
  open(unit=inpunit,file='frame.inp',status='old')
  read(inpunit,*) logunit ! determines output type: 2 -> frame.log, 6 -> screen
  read(inpunit,*) nvar    ! number of dependent variables (ODEs)
  read(inpunit,*) maxstep ! (maximum) number of steps
  read(inpunit,*) kmax    ! 0: no intermediate storage of results,
                          ! > 0: intermediate steps stored in arrays xp and yp.
  read(inpunit,*) iwhich  ! integrator switch:
                          !   0: Euler step
                          !   1: RK 4th order with fixed step size
                          !   2: RK 5th order with adaptive stepsize
                          !   3: Kapps-Rosenbrock semi-implicit
  read(inpunit,*) iprob   ! problem number:
                          !   1: y' = a y
                          !   2: stiff equation (1)
                          !   3: stiff equation (2)
  read(inpunit,*) x1      ! lower boundary
  read(inpunit,*) x2      ! upper boundary
  read(inpunit,*) eps     ! accuracy (for rkqs and stiff)
  read(inpunit,*) hmin    ! minimum allowed step size
  read(inpunit,*) h1      ! first guess at step size.
  read(inpunit,*) dxsav   ! step interval for storing intermediate results. If 0.0, store each step.
  read(inpunit,*) label   ! model specifier with 6 characters
  close(inpunit)

  if (iprob .ne. 1 .or. iwhich .gt. 1) stop ' wrong input for frame_error'

  if (kmax .ne. 1) stop ' set kmax = 1'

  if (logunit .eq. 2) then
    open(unit=logunit,file='frame.log',status='unknown')
  endif

  dxsav_orig = dxsav

  if(dxsav .gt. 0.d0) then
    i = dabs(x2-x1)/dxsav+1
    if(kmax .lt. i) then
      write(logunit,'("FRAME: Warning!!! kmax might be too small")')
      write(logunit,'("FRAME: kmax should be increased to ",i0)') i
    endif
  endif

! set the necessary array sizes depending on the problem.
! Needs to be modified for a new problem.
  call setproblem(iprob, nvar, narg, logunit)

! allocate grids
  allocate(xp    (kmax))
  allocate(yp    (kmax,1:nvar))
  allocate(ystart(1:nvar))
  allocate(argarr(1:narg))

  write(fnm,'("ode_",a6,".dmp")') label
  open(unit=dmpunit,file=fnm,status="unknown") ! output_file

  if (iwhich.eq.0) stepfac = 10
  if (iwhich.eq.1) stepfac =  2.5

  do it=0,10

    dxsav = dxsav_orig
    if (it .gt. 0) maxstep = maxstep*stepfac
    write(0,*) 'maxstep = ',maxstep
    h = abs(x2-x1)/dble(maxstep)

! initialize ystart, arrarg, etc.
! Needs to be modified for a new problem.
    call initproblem(iprob, nvar, narg, ystart, argarr, inpunit, logunit, x1)

    nok  =  0
    nbad =  0
    done = -1
! Now call the integrator (strictly: the driver for the integrator)
    call odeint (nvar   , maxstep, kmax  , kcount, iwhich ,          &
                 x1     , x2     , eps   , h1    , hmin   , dxsav  , &
                 nok    , nbad   ,                                   &
                 ystart ,                                            &
                 narg   , argarr ,                                   &
                 xp     , yp     ,                                   &
                 done   , logunit)

! diagnostics
    write(0,'(a,i0,3es24.16)') 'maxstep,x2,xp,yp = ',maxstep,x2,xp(1),yp(1,1)
    write(logunit,'("FRAME: odeint terminated with status ",i2)') done
    write(logunit,'("FRAME: nok  = ",i12)') nok
    write(logunit,'("FRAME: nbad = ",i12)') nbad

! save results, if successful
!   if (done .ne. 1) then
    if (done .eq. 2) then
      write(logunit,'("FRAME: no success, no file to be written")')
      goto 20
    endif

    write(logunit,'("FRAME: saving file ",a15)') fnm
    if (kcount .ne. 1) stop ' something wrong in frame_error'
    yexact = 1.d-3*(exp(2.5d0*x2))
    do i=1,kcount
      write(dmpunit,10) it, h, yp(i,:), (yp(i,:) - yexact)/yexact
    enddo

  enddo

  close(dmpunit)
10 format(i3,2x,3(e22.16,2x)) ! if nvar > 3, adapt descriptor

20 deallocate(ystart)
  deallocate(xp)
  deallocate(yp)
  deallocate(argarr)

end
!=======================================================================

