!=======================================================================
!
! ROUTINE    : nr_ludcmp
!
! PURPOSE    : Replaces matrix a(1:n,1:n) by LU-decomposition
!              of row-wise permutation of itself.
!              Used in combination with nr_lubksb.
!
! PARAMETERS : n   : dimension
!              a   : matrix (input and output)
!              indx: on leaving, contains row permutations
!              d   : +1: even permutation
!                    -1: odd  permutation
!              vtmp: scratch vector
!              nr_err: 0 if no error occured, 1 if matrix is singular
!
!=======================================================================

subroutine ludcmp (n, a, indx, d, vtmp, logunit, nr_err)

  implicit none

! subroutine arguments
  integer          :: n, logunit, nr_err
  integer          :: indx(1:n)
  double precision :: d
  double precision :: vtmp(1:n)
  double precision :: a   (1:n,1:n)

! local variables
  integer          :: i, imax, j, k
  double precision :: big, dum, sumtmp, temp

  double precision, parameter :: tiny = 1.0d-30

!-----------------------------------------------------------------

  nr_err = 0

  d = 1.0d0
  do i = 1, n
    big = maxval(dabs(a(i,:)))
    if (big .eq. 0.d0) then
      write(logunit,'("NR_LUDCMP: singular matrix.")')
      nr_err = 1
      return
    endif
    vtmp(i) = 1.d0 / big
  enddo

  do j = 1, n
    do i = 1, j-1
      sumtmp = a(i,j)
      do k = 1, i-1
        sumtmp = sumtmp - a(i,k) * a(k,j)
      enddo
      a(i,j) = sumtmp
    enddo

    big = 0.d0
    do i = j, n
      sumtmp = a(i,j)
      do k = 1, j-1
        sumtmp = sumtmp - a(i,k) * a(k,j)
      enddo
      a(i,j) = sumtmp
      dum    = vtmp(i)*dabs(sumtmp)
      if (dum .ge. big) then
        big  = dum
        imax = i
      endif
    enddo

    if (j .ne. imax) then
      do k = 1, n
        dum       = a(imax,k)
        a(imax,k) = a(j,k)
        a(j,k)    = dum
      enddo
      d          = -d
      vtmp(imax) = vtmp(j)
    endif

    indx(j) = imax
    if (a(j,j) .eq. 0.d0) then
      a(j,j) = tiny
      write(logunit,&
            '("NR_LUDCMP: warning: replaced a(",i4,i4,") with tiny")') j,j
    endif
    if (j .ne. n) then
      dum = 1.d0 / a(j,j)
      do i = j+1, n
        a(i,j) = a(i,j) * dum
      enddo
    endif
  enddo

  return
end

!=======================================================================

