;==============================================================================
; PROGRAM: plotode.pro
;
; PURPOSE: plots results of frame.F/odeint.F
;
; ARGUMENTS: iprob    : no of problem, 1 to 3
;            models   : array of model numbers to plot, e.g.,
;                         [2]     = plot only model 2
;                         [3,1,4] = plot models 3, 1, and 4
; KEYWORDS:  /solu    : overplot analytical solution (for problems 1 and 2)
;                       and determine mean quadratic deviation
;            /points  : plot points instead of lines (use with solu)
;            /logy    : plot the y-axis logarithmically
;            /logerr  : plot the errors logarithmically (for problems 1 and 2)
;            /color   : plot data using color instead of b/w
;            /print   : output to postscript file plotode_....ps
;            ps="..." : extra string for the postscript filename
;
;==============================================================================

pro plotode,iprob,models,solu=solu,points=points,print=print,ps=ps,$
            logy=logy,logerr=logerr,color=color, $
            xrange=xrange,yrange=yrange,ctype=ctype

  nmod = n_elements(models)
  models = models-1 ; IDL uses 0-based indexing, but we number our models from 1

; if required, adapt following choices:
;------------------------------------------------------------------------------
  run1   = ['problem1','problem1','problem1','problem1','problem1','problem1']
  run2   = ['problem2','problem2','problem2','problem2','problem2','problem2']
  run3   = ['problem3','problem3']

  label1 = ['00p1i1','00p1i2','00p1i3','00p1i4','00p1i5','00p1i6']
  label2 = ['00p2i1','00p2i2','00p2i3','00p2i4','00p2i5','00p2i6']
  label3 = ['00p3i1','00p3i2']
;------------------------------------------------------------------------------

; Process arguments and options.
  if iprob lt 0 or iprob gt 3 then begin
    print,' Error: wrong input of iprob, try again!'
    return
  endif

  if keyword_set(solu) then begin
    if iprob ne 1 and iprob ne 2 then begin
      print,' Error: solution available only for iprob = 1 or iprob = 2, try again!'
      return
    endif
  endif

  compile_opt strictarrsubs ; disallow out-of-range subscripts

  case iprob of
    1: run = run1[models]
    2: run = run2[models]
    3: run = run3[models]
  endcase

  case iprob of
    1: label = label1[models]
    2: label = label2[models]
    3: label = label3[models]
  endcase

  if not keyword_set(logy)   then logy = 0
  if not keyword_set(logerr) then logerr = 0
  if not keyword_set(color)  then color = 0

  if color then begin
    if keyword_set(print) then begin  ; white background
      tvlct,255*[0,1,0, 0,0, 1,.8,.8],$ ; red
            255*[0,0,.9,0,.9,0,.8,.8],$ ; green
            255*[0,0,0, 1,.9,1,0 ,.8]   ; blue
    endif else begin                  ; black background
      tvlct,255*[0,1,0,0, 0,1,1,1],$    ; red
            255*[0,0,1,.3,1,0,1,1],$    ; green
            255*[0,0,0,1, 1,1,0,1]      ; blue
    endelse
    usersym,[0,1,0,-1,0],[1,0,-1,0,1],/fill
    ptype = [8,8,8,8,8,8,8] ; point types
    ltype = [0,0,0,0,0,0,0] ; line styles
    if not keyword_set(ctype) then ctype = [1,2,3,4,5,6,7] ; colors
  endif else begin
    if keyword_set(print) then tvlct,[0],[0],[0],1 else tvlct,[255],[255],[255],1
    ptype = [1,2,4,5,6,7,1] ; point types
    ltype = [0,2,3,4,5,6,1] ; line styles
    ctype = [1,1,1,1,1,1,1] ; colors
  endelse

  descr = ''
  if keyword_set(ps) then descr = '_'+ps

  disk = '../' ; path to model directories

;------------------------------------------------------------------------------

  xarr  = ptrarr(nmod)
  yarr  = ptrarr(nmod)
  sarr  = ptrarr(nmod)
  rarr  = ptrarr(nmod)
  minx  = dblarr(nmod)
  maxx  = dblarr(nmod)
  minya = ptrarr(nmod)
  maxya = ptrarr(nmod)
  minsa = ptrarr(nmod)
  maxsa = ptrarr(nmod)
  minra = ptrarr(nmod)
  maxra = ptrarr(nmod)

; Load the models from the data files.
  kcount = 0l
  nvar   = 0l
  for m=0,nmod-1 do begin
    fnm = disk+'run/'+run[m]+'/ode_'+label[m]+'.dmp'
    openr,1,fnm
    readf,1,kcount,nvar
    print,format='("PLOTODE: kcount = ",i8," nvar = ",i2)',kcount,nvar
;   xarr [m] = ptr_new(dblarr(kcount))
;   yarr [m] = ptr_new(dblarr(kcount,nvar))
    xyvec = dblarr(nvar+1,kcount)
    readf,1,xyvec
    close,1
    xyvec=transpose(xyvec)
    xarr [m] = ptr_new(xyvec(*,0))
    yarr [m] = ptr_new(xyvec(*,1:nvar))
    minya[m] = ptr_new(dblarr(nvar))
    maxya[m] = ptr_new(dblarr(nvar))
    minx[m] = min((*xarr[m]))
    maxx[m] = max((*xarr[m]))

; Also calculate errors for the models, if analytic solution is known.
    if keyword_set(solu) then begin
      sarr [m] = ptr_new(dblarr(kcount,nvar))
      minsa[m] = ptr_new(dblarr(nvar))
      maxsa[m] = ptr_new(dblarr(nvar))
      rarr [m] = ptr_new(dblarr(kcount,nvar))
      minra[m] = ptr_new(dblarr(nvar))
      maxra[m] = ptr_new(dblarr(nvar))
      if (iprob eq 1) then begin
        (*sarr[m])[*,0] = (1d-3)*exp(2.5d0*(*xarr[m]))
      endif
      if (iprob eq 2) then begin
        (*sarr[m])[*,0] = 2d0*exp(-(*xarr[m]))-exp(-1.0d3*(*xarr[m]))
        (*sarr[m])[*,1] =    -exp(-(*xarr[m]))+exp(-1.0d3*(*xarr[m]))
      endif
      for n=0,nvar-1 do begin
        (*rarr[m])[*,n] = ((*yarr[m])[*,n]-(*sarr[m])[*,n])/((*sarr[m])[*,n]+1d-30)
        print,models[m],n+1,sqrt(mean((*rarr[m])[*,n]^2)),$
          format='("PLOTODE: mean rel err ",i4,i4," : ",e23.15)'
      endfor
      for n=0,nvar-1 do begin
        if logerr then begin
          *rarr[m] = abs(*rarr[m])
          idx = where((*rarr[m])[*,n] ne 0.)
          (*minra[m])[n] = min((*rarr[m])[idx,n])
          (*maxra[m])[n] = max((*rarr[m])[idx,n])
        endif else begin
          (*minra[m])[n] = min((*rarr[m])[*,n])
          (*maxra[m])[n] = max((*rarr[m])[*,n])
        endelse
        if logy then begin
          *sarr[m] = abs(*sarr[m])
          idx = where((*sarr[m])[*,n] ne 0.)
          (*minsa[m])[n] = min((*sarr[m])[idx,n])
          (*maxsa[m])[n] = max((*sarr[m])[idx,n])
        endif else begin
          (*minsa[m])[n] = min((*sarr[m])[*,n])
          (*maxsa[m])[n] = max((*sarr[m])[*,n])
        endelse
      endfor
    endif

    for n=0,nvar-1 do begin
      if logy then begin
        *yarr[m] = abs(*yarr[m])
        idx = where((*yarr[m])[*,n] ne 0.)
        (*minya[m])[n] = min((*yarr[m])[idx,n])
        (*maxya[m])[n] = max((*yarr[m])[idx,n])
      endif else begin
        (*minya[m])[n] = min((*yarr[m])[*,n])
        (*maxya[m])[n] = max((*yarr[m])[*,n])
      endelse
    endfor
  endfor

; Find maximum and minimum values over all models.
  minx = min(minx)
  maxx = max(maxx)
  miny = dblarr(nvar)
  maxy = dblarr(nvar)
  minr = dblarr(nvar)
  maxr = dblarr(nvar)
  miny[*] =  1d30
  maxy[*] = -1d30
  minr[*] =  1d30
  maxr[*] = -1d30
  for m=0,nmod-1 do begin
    for n=0,nvar-1 do begin
      if keyword_set(solu) then begin
        miny[n] = min([miny[n],(*minya[m])[n],(*minsa[m])[n]])
        maxy[n] = max([maxy[n],(*maxya[m])[n],(*maxsa[m])[n]])
        minr[n] = min([minr[n],(*minra[m])[n]])
        maxr[n] = max([maxr[n],(*maxra[m])[n]])
      endif else begin
        miny[n] = min([miny[n],(*minya[m])[n]])
        maxy[n] = max([maxy[n],(*maxya[m])[n]])
      endelse
    endfor
  endfor

; Create separate arrays for plotting the analytic solution,
; with a fixed resolution independent of the model stepsize(s).
  if keyword_set(solu) then begin
    if (iprob eq 1) then begin
      solx = dindgen(1001)*(maxx-minx)/1000.+minx
      soly = dblarr(1001,1)
      soly[*,0] = (1d-3)*exp(2.5d0*solx)
    endif
    if (iprob eq 2) then begin
      solx = dindgen(10001)*(maxx-minx)/10000.+minx
      soly = dblarr(10001,2)
      soly[*,0] = 2d0*exp(-solx)-exp(-1.0d3*solx)
      soly[*,1] =    -exp(-solx)+exp(-1.0d3*solx)
    endif
    if logy then soly = abs(soly)
  endif

; print,minx,maxx,format='("PLOTODE: xmin    = ",e13.5," xmax    = ",e13.5)'
; for n=0,nvar-1 do begin
;   print,n+1,miny[n],n+1,maxy[n],$
;     format='("PLOTODE: ymin[",i2,"]= ",e13.5," ymax[",i2,"]= ",e13.5)'
; endfor

  if keyword_set(xrange) then begin
    minx = xrange[0]
    maxx = xrange[1]
  endif

  if keyword_set(yrange) then begin
    miny = yrange[0,*]
    maxy = yrange[1,*]
  endif

; Select suitable axis ranges for plotting.
  dx = maxx-minx
  minx = minx-0.05*dx
  maxx = maxx+0.05*dx
  if logy then begin
    dy = maxy/miny
    miny = miny/dy^0.05
    maxy = maxy*dy^0.05
    ytitlef = '("!9!! !6!8y!6!D",i0,"!N!9 !!")'
    filext = '_log'
  endif else begin
    dy = maxy-miny
    miny = miny-0.05*dy
    maxy = maxy+0.05*dy
    ytitlef = '("!8y!6!D",i0,"!N")'
    filext = '_lin'
  endelse
  if logerr then begin
    dr = maxr/minr
    minr = minr/dr^0.05
    maxr = maxr*dr^0.05
    ytitlee = '("!9!! !6!7D!8y!6!D",i0,"!N/!8y!6!D",i0,"!N!9 !!")'
    filerr = '_log'
  endif else begin
    dr = maxr-minr
    minr = minr-0.05*dr
    maxr = maxr+0.05*dr
    ytitlee = '("!7D!8y!6!D",i0,"!N/!8y!6!D",i0,"!N")'
    filerr = '_lin'
  endelse

; Set up nvar plots in one column.
  !p.multi=[0,1,nvar]
  !p.font = -1

  if keyword_set(print) then begin
    set_plot,'ps'
    device,filename=disk+'plots/plotode'+filext+descr+'.ps',/color,$
      xsize=19,ysize=20.*(1.-exp(-nvar)),xoffset=1,yoffset=1
    if nvar ge 3 then !p.charsize=2.4 else !p.charsize = 1.2
    symsize=0.7
  endif else begin
    Device, Window_State=windowstate
    if windowstate[0] then wset,0 else window,0,xsize=600,ysize=min([300*nvar,900])
    !p.charsize=1.5
    symsize=1
  endelse

; Plot y values for all models, and analytic solution if avaliable.
  for n=0,nvar-1 do begin
    plot,[minx,maxx],[miny[n],maxy[n]],/nodata,xstyle=1,ystyle=1,ylog=logy,$
      font=-6,xtitle='!8x!6',ytitle=string(n+1,format=ytitlef),charthick=1.5,title='!6'
    for m=0,nmod-1 do begin
      if keyword_set(points) then begin
        oplot,(*xarr[m]),(*yarr[m])[*,n],psym=ptype[m],color=ctype[m],symsize=symsize,thick=2
        if not keyword_set(solu) then $
          oplot,(*xarr[m]),(*yarr[m])[*,n],line=ltype[m],color=ctype[m]
      endif else begin
        oplot,(*xarr[m]),(*yarr[m])[*,n],line=ltype[m],color=ctype[m],thick=2
      endelse
;     if keyword_set(solu) then oplot,(*xarr[m]),(*sarr[m])[*,n],line=ltype[m]
    endfor
    if keyword_set(solu) then oplot,solx,soly[*,n],line=0
  endfor

  if keyword_set(print) then begin
    device,/close
    set_plot,'x'
  endif

  if keyword_set(solu) then begin

    if keyword_set(print) then begin
      set_plot,'ps'
      device,filename=disk+'plots/plotode_err'+filerr+descr+'.ps',/color,$
        xsize=19,ysize=20.*(1.-exp(-nvar)),xoffset=1,yoffset=1
      if nvar ge 3 then !p.charsize=2.4 else !p.charsize = 1.2
      symsize=0.7
    endif else begin
      if windowstate[1] then wset,1 else window,1,xsize=600,ysize=min([300*nvar,900])
      !p.charsize=1.5
      symsize=1
    endelse

; Also plot y value errors for all models, if analytic solution is available.
    for n=0,nvar-1 do begin
      plot,[minx,maxx],[minr[n],maxr[n]],/nodata,xstyle=1,ystyle=1,ylog=logerr,$
        font=-6,xtitle='!8x!6',ytitle=string(n+1,n+1,format=ytitlee),charthick=1.5,title='!6'
      for m=0,nmod-1 do begin
        if keyword_set(points) then begin
          oplot,(*xarr[m]),(*rarr[m])[*,n],psym=ptype[m],color=ctype[m],symsize=symsize,thick=2
          oplot,(*xarr[m]),(*rarr[m])[*,n],line=ltype[m],color=ctype[m]
        endif else begin
          oplot,(*xarr[m]),(*rarr[m])[*,n],line=ltype[m],color=ctype[m],thick=2
        endelse
      endfor
    endfor

    if keyword_set(print) then begin
      device,/close
      set_plot,'x'
    endif

  endif

  !p.multi=0

  return
end

;==============================================================================

