!=======================================================================
!
! SUBROUTINE: rk4.F
!
! PURPOSE   : Performs one RK-step.
!
! SOURCE    : Numerical Recipes, Press et al. 2nd ed. 1992
!
! INPUT     : nvar   : number of ODEs
!             x      : current x (independent variable)
!             h      : step size over which to advance the solution
!             y      : current values of ODEs
!             dydx   : derivatives at x (and work array)
!             narg   : number of problem-specific arguments for derivs
!             argarr : problem-specific coefficients for derivs
!             logunit
!
! OUTPUT    : yout   : incremented values of ODEs (new solution)
!
! EXTERNAL  : subroutine derivs(nvar, x, y, narg, argarr, dxdy)
!
! USAGE     : Given y at x, advances solution by step size h using the
!             derivatives in the user-supplied subroutine derivs.
!
!=======================================================================

subroutine rk4 (nvar   ,                &
                x      , h      ,       &
                y      , dydx   ,       &
                narg   , argarr ,       &
                yout   ,                &
                logunit)

  implicit none

! subroutine arguments
  integer          :: nvar, narg, logunit
  double precision :: x, h
  double precision :: y     (nvar)
  double precision :: argarr(narg)
  double precision :: dydx  (nvar)
  double precision :: yout  (nvar)

! local variables
  integer                       :: i
  double precision              :: xh, hh, h6
  double precision, allocatable :: dym(:)
  double precision, allocatable :: dyt(:)
  double precision, allocatable :: yt (:)

!-----------------------------------------------------------------

! allocate local vectors
  allocate(dym(nvar))
  allocate(dyt(nvar))
  allocate(yt (nvar))

  hh = 0.5d0 * h
  h6 = h / 6.d0
  xh = x + hh

! Note: for the current x, odeint has already set dydx by a call to derivs.
  yt (:) = y  (:) + dydx(:) * hh                ! first step
  call derivs(nvar, xh, yt, narg, argarr, dyt)  ! second step
  yt (:) = y  (:) + dyt (:) * hh
  call derivs(nvar, xh, yt, narg, argarr, dym)  ! third step
  yt (:) = y  (:) + dym (:) * h
  dym(:) = dym(:) + dyt (:)
  call derivs(nvar, x+h, yt, narg, argarr, dyt) ! fourth step

  yout(:) = y(:) + (dydx(:)+dyt(:)+2.d0*dym(:))*h6

  x       = x + h

  deallocate(dym)
  deallocate(dyt)
  deallocate(yt)

  return
end

!=======================================================================

