!=======================================================================
!
! SUBROUTINE: stiff.F
!
! PURPOSE   : Fourth-order Rosenbrock step integrating stiff ODEs with
!             adaptive stepsize control.
!
! SOURCE    : Numerical Recipes, Press et al. 2nd ed. 1992
!
! INPUT     : nvar   : number of ODEs
!             x      : current x (independent variable). See OUTPUT.
!             htry   : step size over which to advance the solution (attempt)
!             eps    : desired accuracy ("Delta_0")
!             y      : current values of ODEs
!             dydx   : derivatives at x (and work array)
!             yscal  : scaling array for errors
!             narg   : number of problem-specific arguments for derivs
!             argarr : problem-specific coefficients for derivs
!             logunit
!
! OUTPUT    : x      : advanced x (by hdid)
!             y      : incremented values of ODEs (new solution)
!             hdid   : actual step size used
!             hnext  : estimated next step size
!
! EXTERNAL  : subroutine derivs  (nvar, x, y, narg, argarr, dxdy)
!             subroutine jacobian(nvar, x, y, narg, argarr, dfdx, dfdy)
!             subroutine ludcmp
!             subroutine lubksb
!
! USAGE     : Given y at x, advances solution by step size h using the
!             derivatives in the user-supplied subroutine derivs.
!
!=======================================================================

subroutine stiff(nvar   ,                  &
                 x      , htry   , eps   , &
                 y      , dydx   , yscal , &
                 narg   , argarr ,         &
                 hdid   , hnext  , &
                 logunit)

  implicit none

! subroutine arguments
  integer          :: nvar, narg, logunit, ierr
  double precision :: x, htry, eps, hdid, hnext
  double precision :: y     (nvar)
  double precision :: dydx  (nvar)
  double precision :: argarr(narg)
  double precision :: yscal (nvar)

! local variables
  logical          :: done
  integer          :: i, j, jtry
  double precision :: d, errmax, h, xsav

  integer,          allocatable, dimension(:)   :: indx
  double precision, allocatable, dimension(:)   :: dfdx, dysav, ysav, yerr, &
                                                   vtmp, g1, g2, g3, g4
  double precision, allocatable, dimension(:,:) :: dfdy, a

  integer,          parameter :: maxtry = 40
  double precision, parameter :: safety = 0.9d0, errcon = 0.1296d0
  double precision, parameter :: grow   = 1.5d0, pgrow  = -0.25d0
  double precision, parameter :: shrnk  = 0.5d0, pshrnk = -1.d0/3.d0

! Parameters from Shampine:
! double precision, parameter :: &
!   gam =   0.5,       &
!   a21 =   2.0,       &
!   a31 =   1.92,      &
!   a32 =   0.24,      &
!   c21 =  -8.0,       &
!   c31 =  14.88,      &
!   c32 =   2.4,       &
!   c41 =  -0.896,     &
!   c42 =  -0.432,     &
!   c43 =  -0.4,       &
!   b1  =  19.0/9.0,   &
!   b2  =   0.5,       &
!   b3  =  25.0/108.0, &
!   b4  = 125.0/108.0, &
!   e1  =  17.0/54.0,  &
!   e2  =   7.0/36.0,  &
!   e3  =   0.0,       &
!   e4  = 125.0/108.0, &
!   c1x =   0.5,       &
!   c2x =  -1.5,       &
!   c3x =   2.42,      &
!   c4x =   0.116,     &
!   a2x =   1.0,       &
!   a3x =   0.6

! Parameters from Kaps & Rentrop:
  double precision, parameter :: &
    gam =  0.231d0,           &
    a21 =  2.0d0,             &
    a31 =  4.52470820736d0,   &
    a32 =  4.16352878860d0,   &
    c21 = -5.07167533877d0,   &
    c31 =  6.02015272865d0,   &
    c32 =  0.159750684673d0,  &
    c41 = -1.856343618677d0,  &
    c42 = -8.50538085819d0,   &
    c43 = -2.08407513602d0,   &
    b1  =  3.95750374663d0,   &
    b2  =  4.62489238836d0,   &
    b3  =  0.617477263873d0,  &
    b4  =  1.282612945268d0,  &
    e1  = -2.30215540292d0,   &
    e2  = -3.07363448539d0,   &
    e3  =  0.873280801802d0,  &
    e4  =  1.282612945268d0,  &
    c1x =  gam,               &
    c2x = -0.396296677520d-1, &
    c3x =  0.550778939579d0,  &
    c4x = -0.553509845700d-1, &
    a2x =  0.462d0,           &
    a3x =  0.880208333333d0

!-----------------------------------------------------------------


! allocate local vectors
  allocate(indx (nvar))
  allocate(dfdx (nvar))
  allocate(vtmp (nvar))
  allocate(dysav(nvar))
  allocate(ysav (nvar))
  allocate(yerr (nvar))
  allocate(g1   (nvar))
  allocate(g2   (nvar))
  allocate(g3   (nvar))
  allocate(g4   (nvar))
  allocate(dfdy (nvar,nvar))
  allocate(a    (nvar,nvar))

! save initial values
  xsav     = x
  ysav (:) = y(:)
  dysav(:) = dydx(:)

  call jacobian(nvar, xsav, ysav, narg, argarr, dfdx, dfdy)

  h    = htry
  done = .false.
  jtry = 0
  do while (.not. done)
    jtry = jtry + 1
    do j=1,nvar
      do i=1,nvar
        a(i,j) = -dfdy(i,j)
      enddo
      a(j,j) = a(j,j) + 1d0/(gam*h)
    enddo
    call ludcmp (nvar, a, indx, d, vtmp, logunit, ierr)
    g1(:) = dysav(:) + h * c1x * dfdx(:)
    call lubksb (nvar, a, indx, g1)
    y(:) = ysav(:) + a21 * g1(:)
    x    = xsav + a2x * h
    call derivs(nvar, x, y, narg, argarr, dydx)
    g2(:) = dydx(:) + h * c2x * dfdx(:) + c21 * g1(:) / h
    call lubksb (nvar, a, indx, g2)
    y(:) = ysav(:) + a31 * g1(:) + a32 * g2(:)
    x    = xsav + a3x * h
    call derivs(nvar, x, y, narg, argarr, dydx)
    g3(:) = dydx(:) + h * c3x * dfdx(:) + (c31 * g1(:) + c32 * g2(:)) / h
    call lubksb (nvar, a, indx, g3)
    g4(:) = dydx(:) + h * c4x * dfdx(:) + (c41 * g1(:) + c42 * g2(:) + c43 * g3(:)) / h
    call lubksb (nvar, a, indx, g4)
    y   (:) = ysav(:) + b1*g1(:) + b2*g2(:) + b3*g3(:) + b4*g4(:)
    yerr(:) =           e1*g1(:) + e2*g2(:) + e3*g3(:) + e4*g4(:)
    x       = xsav + h
    if (x .eq. xsav) then
      write(logunit,'("STIFF: stepsize too small. x,h=",e23.15,e23.15)') x, h
      stop
    endif
    errmax = 0d0 !   evaluate accuracy
    do i = 1, nvar
      errmax = dmax1(errmax,dabs(yerr(i)/yscal(i)))
    enddo
    errmax = errmax / eps ! scale relative to required tolerance
    if (errmax .le. 1d0) then ! success
      done = .true.
      hdid = h
      if (errmax .gt. errcon) then
        hnext = safety * h * errmax**pgrow
      else
        hnext = grow*h
      endif
      done = .true.
    else
      hnext = safety * h * errmax**pshrnk
      if (h .ge. 0.d0) then
        h = dmax1(hnext,shrnk*h)
      else
        h = dmin1(hnext,shrnk*h)
      endif
    endif
    if (jtry .ge. maxtry) then
      write(logunit,'("STIFF: exceeded maxtry")')
      stop
    endif
  enddo

  deallocate(indx)
  deallocate(dfdx)
  deallocate(vtmp)
  deallocate(dysav)
  deallocate(ysav)
  deallocate(yerr)
  deallocate(g1)
  deallocate(g2)
  deallocate(g3)
  deallocate(g4)
  deallocate(dfdy)
  deallocate(a)

  return
end

!=======================================================================

